from .model import Submodel, SubmodelElement, Property, DataTypeDefXsd, MultiLanguageProperty, SubmodelElementCollection
from .adapter import AdapterPath

from aas_test_engines.result import AasTestResult, Level
from typing import List, Dict, Optional, Tuple, Union
from collections import defaultdict
from dataclasses import is_dataclass, fields
from enum import Enum

INVALID = object()
LangString = Dict[str, str]


def _unwrap(cls) -> Tuple[int, int, type]:
    try:
        origin = cls.__origin__
    except AttributeError:
        # neither List nor Optional
        return 1, 1, cls
    # List[T]
    if origin is list:
        return 1, float('inf'), cls.__args__[0]
    # Dict[T] aka LangString
    if origin is dict:
        return 1, 1, LangString

    # We only support Optional aka. Union[x, NoneType]
    assert origin is Union, origin
    assert len(cls.__args__) == 2 and cls.__args__[1] is type(None)
    cls = cls.__args__[0]
    try:
        origin = cls.__origin__
    except AttributeError:
        # Optional[T]
        return 0, 1, cls
    # Optional[Dict[T]] aka Optional[LangString]
    if origin is dict:
        return 0, 1, LangString
    assert origin is list
    # Optional[List[T]]
    return 0, float('inf'), cls.__args__[0]


def _parse_string(root_result: AasTestResult, element: SubmodelElement, path: AdapterPath) -> str:
    if not isinstance(element, Property):
        root_result.append(AasTestResult(f"Cannot convert to string: must be a Property @ {path}", level=Level.ERROR))
        return
    if element.value_type != DataTypeDefXsd.string:
        root_result.append(AasTestResult(f"Cannot convert to string: valueType must be xs:string @ {path}", level=Level.ERROR))
        return
    if not element.value:
        root_result.append(AasTestResult(f"Cannot convert to string: no value @ {path}", level=Level.ERROR))
        return
    return element.value.raw_value


def _parse_lang_string(root_result: AasTestResult, element: SubmodelElement, path: AdapterPath) -> LangString:
    if not isinstance(element, MultiLanguageProperty):
        root_result.append(AasTestResult(f"Cannot convert to lang string: not a MultiLanguageProperty @ {path}"))
        return
    if not element.value:
        root_result.append(AasTestResult(f"Cannot convert to lang string: no value @ {path}"))
        return
    result: LangString = {}
    for i in element.value:
        result[i.language] = i.text
    return result


def _parse_element_collection(root_result: AasTestResult, cls, collection: SubmodelElementCollection, path: AdapterPath):
    if collection.value is None:
        root_result.append(AasTestResult(f"No elements @ {path}", level=Level.ERROR))
        return INVALID
    return _parse_elements(root_result, cls, collection.value, path)


def _parse_elements(root_result: AasTestResult, cls, elements: List[SubmodelElement], path: AdapterPath):
    elements_by_semantic_id: Dict[str, List[SubmodelElement]] = defaultdict(list)
    for collection in elements:
        if collection.semantic_id:
            key = collection.semantic_id.keys[0].value.raw_value
            elements_by_semantic_id[key].append(collection)
    args = {}
    all_semantic_ids = set()
    for field in fields(cls):
        try:
            semantic_id = field.metadata['semantic_id']
        except KeyError:
            raise Exception(f"Field {field.name} is missing metadata 'semantic_id'")
        if semantic_id in all_semantic_ids:
            raise Exception(f"Duplicate metadata semantic_id {semantic_id}")
        all_semantic_ids.add(semantic_id)

        min_card, max_card, field_type = _unwrap(field.type)
        sub_elements = elements_by_semantic_id[semantic_id]
        if len(sub_elements) > max_card:
            root_result.append(AasTestResult(f"Field {field.name}: found {len(sub_elements)} elements with semanticId {semantic_id}, but at most {max_card} allowed", level=Level.ERROR))
            field_val = INVALID
        elif len(sub_elements) < min_card:
            root_result.append(AasTestResult(f"Field {field.name}: found {len(sub_elements)} elements with semanticId {semantic_id}, but at least {min_card} required", level=Level.ERROR))
            field_val = INVALID
        else:
            if len(sub_elements) == 0:
                field_val = None
            elif max_card == 1:
                field_val = _parse(field_type, sub_elements[0], root_result, path + field.name)
            else: # i.e. list
                field_val = [_parse(field_type, el, root_result, path + field.name + idx) for idx, el in enumerate(sub_elements)]
        args[field.name] = field_val
    return cls(**args)


def _parse_enum(root_result: AasTestResult, cls, elements: List[SubmodelElement], path: AdapterPath):
    value = _parse_string(root_result, elements, path)
    if not root_result.ok():
        return INVALID
    try:
        return cls(value)
    except ValueError as e:
        root_result.append(AasTestResult(f"{e} @ {path}", level=Level.ERROR))
        return INVALID


def _parse(cls, element: SubmodelElement, root_result, path: AdapterPath):
    origin = getattr(cls, '__origin__', None)
    if origin:
        if origin is dict:
            return _parse_lang_string(root_result, element, path)
    else:
        if is_dataclass(cls):
            return _parse_element_collection(root_result, cls, element, path)
        elif cls is str:
            return _parse_string(root_result, element, path)
        elif isinstance(cls, Enum.__class__):
            return _parse_enum(root_result, cls, element, path)
    raise NotImplementedError(
        f"There is no parsing implemented for:\n"
        f"  origin:    {origin}\n"
        f"  args:      {getattr(cls, '__args__', None)}\n"
        f"  cls:       {cls}\m"
    )


def parse_submodel(root_result: AasTestResult, cls, submodel: Submodel):
    return _parse_elements(root_result, cls, submodel.submodel_elements, AdapterPath())



from typing import Tuple, Optional
from aas_test_engines.result import AasTestResult

from .parse import parse_and_check_json, parse_and_check_xml
from .model import Environment
from .submodel_templates import parse_submodel_templates


def json_to_env(value: any) -> Tuple[AasTestResult, Optional[Environment]]:
    result, env = parse_and_check_json(Environment, value)
    if result.ok():
        parse_submodel_templates(result, env)
    return result, env


def xml_to_env(value: any) -> Tuple[AasTestResult, Optional[Environment]]:
    result, env = parse_and_check_xml(Environment, value)
    if result.ok():
        parse_submodel_templates(result, env)
    return result, env



from unittest import TestCase
import os
import zipfile
import io
import json
from xml.etree import ElementTree

from aas_test_engines import file, exception
from aas_test_engines.result import Level

script_dir = os.path.dirname(os.path.realpath(__file__))


def in_memory_zipfile(path: str):
    buffer = io.BytesIO()
    zip = zipfile.ZipFile(buffer, "a", zipfile.ZIP_DEFLATED, False)
    for root, subdirs, files in os.walk(path):
        for file in files:
            real_path = os.path.join(root, file)
            archive_path = real_path[len(path)+1:]
            zip.write(real_path, archive_path)
    return zip


class CheckJsonTest(TestCase):

    def test_empty(self):
        result = file.check_json_data({})
        self.assertTrue(result.ok())

    def test_no_json(self):
        result = file.check_json_file(io.StringIO("no json"))
        self.assertFalse(result.ok())


class CheckXmlTest(TestCase):

    def test_empty(self):
        data = ElementTree.fromstring(
            """<environment xmlns="https://admin-shell.io/aas/3/0">
            </environment>""")
        result = file.check_xml_data(data)
        self.assertTrue(result.ok())

    def test_invalid_namespace(self):
        data = ElementTree.fromstring(
            """<environment xmlns="invalid">
            </environment>""")
        result = file.check_xml_data(data)
        self.assertFalse(result.ok())
        result.dump()

    def test_no_xml(self):
        result = file.check_xml_file(io.StringIO("no xml"))
        self.assertFalse(result.ok())


class CheckAasxTest(TestCase):

    def test_not_a_zip(self):
        with open(os.path.join(script_dir, 'fixtures/aasx/invalid/invalid_json/[Content_Types].xml'), "rb") as f:
            result = file.check_aasx_file(f)
        result.dump()
        self.assertEqual(result.level, Level.ERROR)

    def test_empty(self):
        z = in_memory_zipfile(os.path.join(
            script_dir, 'fixtures/aasx/invalid/empty'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.ERROR)

    def test_no_rels(self):
        z = in_memory_zipfile(os.path.join(
            script_dir, 'fixtures/aasx/invalid/no_rels'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.ERROR)

    def test_invalid_rels(self):
        z = in_memory_zipfile(os.path.join(
            script_dir, 'fixtures/aasx/invalid/invalid_rels'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.ERROR)

    def test_unknown_filetype(self):
        z = in_memory_zipfile(os.path.join(
            script_dir, 'fixtures/aasx/invalid/unknown_filetype'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.WARNING)

    def test_rel_target_not_exists(self):
        z = in_memory_zipfile(os.path.join(script_dir, 'fixtures/aasx/invalid/rel_target_not_exists'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.ERROR)
        self.assertTrue(any("Relationship has non-existing target aasx/aasx-origin" in i for i in result.to_lines()))

    def test_no_aas(self):
        z = in_memory_zipfile(os.path.join(script_dir, 'fixtures/aasx/valid/no_aas1'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.WARNING)
        z = in_memory_zipfile(os.path.join(script_dir, 'fixtures/aasx/valid/no_aas2'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.WARNING)

    def test_valid_xml(self):
        z = in_memory_zipfile(os.path.join(script_dir, 'fixtures/aasx/valid/xml'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.INFO)

    def test_valid_json(self):
        z = in_memory_zipfile(os.path.join(script_dir, 'fixtures/aasx/valid/json'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.INFO)

    def test_invalid_xml(self):
        z = in_memory_zipfile(os.path.join(script_dir, 'fixtures/aasx/invalid/invalid_xml'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.ERROR)

    def test_invalid_json(self):
        z = in_memory_zipfile(os.path.join(script_dir, 'fixtures/aasx/invalid/invalid_json'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.ERROR)

    def test_relative_paths(self):
        z = in_memory_zipfile(os.path.join(script_dir, 'fixtures/aasx/valid/relative_paths'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.INFO)

    def test_recursive(self):
        z = in_memory_zipfile(os.path.join(script_dir, 'fixtures/aasx/valid/recursive'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertEqual(result.level, Level.WARNING)

    def test_deprecated_rel(self):
        z = in_memory_zipfile(os.path.join(script_dir, 'fixtures/aasx/valid/deprecated_rel'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertTrue(any("Deprecated type http://www.admin-shell.io/" in line for line in result.to_lines()))
        self.assertEqual(result.level, Level.WARNING)


class SupportedVersionTest(TestCase):

    def test_invoke(self):
        s = file.supported_versions()
        for version, templates in s.items():
            print(f"{version}: {', '.join(templates)}")
        self.assertIn(file.latest_version(), s)


class GenerateTest(TestCase):

    def check(self, generator, limit=float('inf')):
        i = 0
        for _ in generator:
            i += 1
            if i > limit:
                break

    def test_meta_model(self):
        generator = file.generate()
        self.check(generator, 10)


class CheckSubmodelTemplate(TestCase):

    def test_contact_info(self):
        with open(os.path.join(script_dir, 'fixtures', 'submodel_templates', 'contact_information.json')) as f:
            data = json.load(f)
        result = file.check_json_data(data)
        result.dump()
        self.assertTrue(result.ok())
        data['submodels'][0]['submodelElements'][0]['value'][0]['value'] = 'invalid'
        result = file.check_json_data(data)
        result.dump()
        self.assertFalse(result.ok())

    def test_digital_nameplate(self):
        with open(os.path.join(script_dir, 'fixtures', 'submodel_templates', 'digital_nameplate.json')) as f:
            data = json.load(f)
        result = file.check_json_data(data)
        result.dump()
        self.assertTrue(result.ok())
        # either family or type must be present
        elements = data['submodels'][0]['submodelElements']
        indices = [idx for idx, value in enumerate(elements) if value['idShort'] in ['ManufacturerProductFamily', 'ManufacturerProductType']]
        for idx in indices:
            del elements[idx]['semanticId']
        result = file.check_json_data(data)
        result.dump()
        self.assertFalse(result.ok())

    def test_no_submodels(self):
        data = {}
        # is compliant to meta-model...
        result = file.check_json_data(data)
        result.dump()
        self.assertTrue(result.ok())



from fences.json_schema.normalize import normalize, NormalizationConfig
from fences.json_schema.parse import default_config, parse, KeyReference
from fences.json_schema.config import FormatSamples
from fences.core.node import Node as FlowGraph, NoOpDecision, Leaf

default_samples = {
    'xs:decimal': FormatSamples(
        valid=["10.4"]
    ),
    "xs:integer": FormatSamples(
        valid=["12"]
    ),
    "xs:float": FormatSamples(
        valid=["12"]
    ),
    "xs:double": FormatSamples(
        valid=["12"]
    ),
    "xs:time": FormatSamples(
        valid=["14:23:00"],
    ),
    "xs:date": FormatSamples(
        valid=["2000-01-01+12:05"]
    ),
    "xs:dateTime": FormatSamples(
        valid=["2000-01-01T14:23:00.66372+14:00"]
    ),
    "xs:dateTimeUTC": FormatSamples(
        valid=["2000-01-01T14:23:00Z"]
    ),
    "xs:gYearMonth": FormatSamples(
        valid=["2000-01+03:00"]
    ),
    "xs:gMonthDay": FormatSamples(
        valid=["--01-01"]
    ),
    "xs:gYear": FormatSamples(
        valid=["2000"]
    ),
    "xs:gMonth": FormatSamples(
        valid=["--04"]
    ),
    "xs:gDay": FormatSamples(
        valid=["---10"],
    ),
    "xs:duration": FormatSamples(
        valid=["-P1Y2M3DT1H"],
    ),
    "xs:byte": FormatSamples(
        valid=["12"]
    ),
    "xs:short": FormatSamples(
        valid=["12"]
    ),
    "xs:int": FormatSamples(
        valid=["12"]
    ),
    "xs:negativeInteger": FormatSamples(
        valid=["-12"]
    ),
    "xs:long": FormatSamples(
        valid=["12"]
    ),
    "xs:nonNegativeInteger": FormatSamples(
        valid=["12"]
    ),
    "xs:nonPositiveInteger": FormatSamples(
        valid=["-12"]
    ),
    "xs:positiveInteger": FormatSamples(
        valid=["12"]
    ),
    "xs:unsignedByte": FormatSamples(
        valid=["12"]
    ),
    "xs:unsignedInt": FormatSamples(
        valid=["12"]
    ),
    "xs:unsignedShort": FormatSamples(
        valid=["12"]
    ),
    "xs:unsignedLong": FormatSamples(
        valid=["12"]
    ),
    "xs:base64Binary": FormatSamples(
        valid=["AA=="]
    ),
    "xs:anyURI": FormatSamples(
        valid=["http://example.com"]
    ),
    "bcpLangString": FormatSamples(
        valid=["de"]
    ),
    'version': FormatSamples(
        valid=['567'],
    ),
    'contentType': FormatSamples(
        valid=["application/json"],
    ),
    'path': FormatSamples(
        valid=["file://example.com/myfile"],
    ),
}

def post_process(entry: dict, result: FlowGraph) -> FlowGraph:
    try:
        check = entry['check']
    except KeyError:
        return result
    sub_root = NoOpDecision(all_transitions=True)
    sub_root.add_transition(result)
    check = entry['check']
    for ch in check:
        if ch == 'Constraint_AASd-124':
            class FixConstraint124(Leaf):
                def apply(self, data: KeyReference) -> any:
                    try:
                        keys = data.get()['keys']
                    except (KeyError, TypeError):
                        return data
                    if not isinstance(keys, list):
                        return data
                    keys.append({
                        'type': 'GlobalReference',
                        'value': 'xyz'
                    })
                    return data
            sub_root.add_transition(FixConstraint124(is_valid=True))
        elif ch == 'Constraint_AASd-125':
            class FixConstraint125(Leaf):
                def apply(self, data: KeyReference) -> any:
                    try:
                        keys: list = data.get()['keys']
                    except (KeyError, TypeError):
                        return data
                    if isinstance(keys, list):
                        for key in keys[1:]:
                            if isinstance(key, dict):
                                key['type'] = 'Range'
                    return data
            sub_root.add_transition(FixConstraint125(is_valid=True))
        elif ch == 'Constraint_AASd-107':
            class FixConstraint107(Leaf):
                def apply(self, data: KeyReference) -> any:
                    l: list = data.get()
                    try:
                        l['semanticIdListElement'] = l['value'][0]['semanticId']
                    except (KeyError, TypeError):
                        pass
                    return data
            sub_root.add_transition(FixConstraint107(is_valid=True))
        elif ch == 'Constraint_AASd-108':
            class FixConstraint108(Leaf):
                def apply(self, data: KeyReference) -> any:
                    l: list = data.get()
                    try:
                        l['typeValueListElement'] = l['value'][0]['modelType']
                    except (KeyError, TypeError):
                        pass
                    return data
            sub_root.add_transition(FixConstraint108(is_valid=True))
        elif ch == 'Constraint_AASd-109':
            class FixConstraint109(Leaf):
                def apply(self, data: KeyReference) -> any:
                    l: list = data.get()
                    try:
                        if l['typeValueListElement'] in ['Property', 'Range']:
                            l['valueTypeListElement'] = l['value'][0]['valueType']
                    except (KeyError, TypeError):
                        pass
                    return data
            sub_root.add_transition(FixConstraint109(is_valid=True))
        elif ch == 'Constraint_AASd-119':
            class FixConstraint119(Leaf):
                def apply(self, data: KeyReference) -> any:
                    d: dict = data.get()
                    try:
                        if any(i.get('kind') == 'TemplateQualifier' for i in d['qualifiers']):
                            d['kind'] = 'Template'
                    except (KeyError, TypeError, AttributeError):
                        pass
                    return data
            sub_root.add_transition(FixConstraint119(is_valid=True))
        elif ch == 'Constraint_AASd-129':
            class FixConstraint129(Leaf):
                def apply(self, data: KeyReference) -> any:
                    d: dict = data.get()
                    try:
                        submodel_elements = d.get('submodelElements', [])
                        for element in submodel_elements:
                            if any(i.get('kind') == 'TemplateQualifier' for i in element['qualifiers']):
                                d['kind'] = 'Template'
                                break
                    except (KeyError, TypeError, AttributeError):
                        pass
                    return data
            sub_root.add_transition(FixConstraint129(is_valid=True))
        elif ch == 'Constraint_AASd-134':
            class FixConstraint134(Leaf):
                def apply(self, data: KeyReference) -> any:
                    d: dict = data.get()
                    try:
                        for idx, var in enumerate(d['inputVariables']):
                            var['value']['idShort'] = f"i{idx}"
                    except (KeyError, TypeError):
                        pass
                    try:
                        for idx, var in enumerate(d['outputVariables']):
                            var['value']['idShort'] = f"o{idx}"
                    except (KeyError, TypeError):
                        pass
                    try:
                        for idx, var in enumerate(d['inoutputVariables']):
                            var['value']['idShort'] = f"io{idx}"
                    except (KeyError, TypeError):
                        pass
                    return data
            sub_root.add_transition(FixConstraint134(is_valid=True))
        elif ch == 'Constraint_AASc-3a-002':
            class FixConstraint008(Leaf):
                def apply(self, data: KeyReference) -> any:
                    d: dict = data.get()
                    try:
                        d['preferredName'] = [{
                            'language': 'en',
                            'text': 'xyz',
                        }]
                    except (KeyError, TypeError, AttributeError) as e:
                        pass
                    return data
            sub_root.add_transition(FixConstraint008(is_valid=True))
        elif ch == 'Constraint_AASc-3a-008':
            class FixConstraint008(Leaf):
                def apply(self, data: KeyReference) -> any:
                    d: dict = data.get()
                    try:
                        if 'value' not in d:
                            d['definition'] = [{
                                'language': 'en',
                                'text': 'xyz',
                            }]
                    except (KeyError, TypeError, AttributeError) as e:
                        pass
                    return data
            sub_root.add_transition(FixConstraint008(is_valid=True))
    return sub_root


def generate_graph(schema) -> FlowGraph:
    norm_config = NormalizationConfig(
        full_merge=False,
        additional_mergers={
            'check': lambda x, y: x + y
        },
        detect_duplicate_subschemas=True,
    )
    schema_norm = normalize(schema, norm_config)
    config = default_config()
    config.normalize = False
    config.format_samples = default_samples
    config.post_processor = post_process
    graph = parse(schema_norm, config)
    return graph



from typing import Dict, List, Union, Optional, Tuple, TypeVar,  Iterator
from aas_test_engines.exception import AasTestToolsException
from aas_test_engines.result import AasTestResult, Level, start, abort, write, ResultException

from fences.open_api.open_api import OpenApi, Operation
from fences.open_api.generate import SampleCache, generate_all, generate_one_valid, Request
from fences.core.util import ConfusionMatrix
from json_schema_tool.schema import parse_schema, ParseConfig, SchemaValidationResult

import os
from yaml import load
try:
    # This one is faster but not available on all systems
    from yaml import CSafeLoader as Loader
except ImportError:
    from yaml import SafeLoader as Loader

from dataclasses import dataclass
import requests

import base64

from .model import AssetAdministrationShell, Environment
from .parse import parse_and_check_json

T = TypeVar("T")


def _first_iterator_item(it: Iterator[T]) -> T:
    return next(iter(it))


def b64urlsafe(value: str) -> str:
    return base64.urlsafe_b64encode(value.encode()).decode()


def _map_error(parent: AasTestResult, error: SchemaValidationResult):
    for i in error.keyword_results:
        if i.ok():
            continue
        kw_result = AasTestResult(i.error_message, Level.ERROR)
        for j in i.sub_schema_results:
            _map_error(kw_result, j)
        parent.append(kw_result)


def _assert(predicate: bool, message, level: Level = Level.ERROR):
    if predicate:
        write(f'{message}: OK')
    else:
        abort(AasTestResult(f'{message}: Fail', level))


def _stringify_path(path: List[Union[str, int]]) -> str:
    return "/".join(str(fragment) for fragment in path)


class _NoDefault:
    pass


def _lookup(value: any, path: List[Union[str, int]], default=_NoDefault, idx: int = 0) -> any:
    if idx >= len(path):
        return value

    fragment = path[idx]
    if isinstance(fragment, str):
        if not isinstance(value, dict):
            if default is _NoDefault:
                abort(f"Cannot look up {_stringify_path(path)}: should be an object")
            else:
                return default
        try:
            sub_value = value[fragment]
        except KeyError:
            if default is _NoDefault:
                abort(f"Cannot look up {_stringify_path(path)}: key '{fragment}' does not exist")
            else:
                return default
    elif isinstance(fragment, int):
        if not isinstance(value, list):
            if default is _NoDefault:
                abort(f"Cannot look up {_stringify_path(path)}: should be an array")
            else:
                return default
        try:
            sub_value = value[fragment]
        except IndexError:
            if default is _NoDefault:
                abort(f"Cannot look up {_stringify_path(path)}: array too short")
            else:
                return default
    return _lookup(sub_value, path, default, idx+1)


def _extend(data: Dict[str, List[str]]) -> dict:
    while True:
        all_resolved = True
        for key, values in data.items():
            new_values = []
            for value in values:
                try:
                    new_values.extend(data[value])
                    all_resolved = False
                except KeyError:
                    new_values.append(value)
            data[key] = new_values
        if all_resolved:
            return data


SSP_PREFIX = "https://admin-shell.io/aas/API/3/0/"

_available_suites = _extend({
    # APIs
    "Asset Administration Shell API": [
        "GetAssetAdministrationShell",
        "PutAssetAdministrationShell",
        "GetAllSubmodelReferences",
        "PostSubmodelReference",
        "DeleteSubmodelReferenceById",
        "GetAssetInformation",
        "PutAssetInformation",
        "GetThumbnail",
        "PutThumbnail",
        "DeleteThumbnail"
    ],
    "Submodel API": [
        "GetSubmodel",
        "GetAllSubmodelElements",
        "GetSubmodelElementByPath",
        "GetFileByPath",
        "PutFileByPath",
        "DeleteFileByPath",
        "PutSubmodel",
        "PatchSubmodel",
        "PostSubmodelElement",
        "PostSubmodelElementByPath",
        "PutSubmodelElementByPath",
        "PatchSubmodelElementByPath",
        # "GetSubmodelElementValueByPath", # TODO
        "DeleteSubmodelElementByPath",
        "InvokeOperationSync_AAS",
        "InvokeOperationAsync_AAS",
        "GetOperationAsyncStatus",
        "GetOperationAsyncResult",
    ],
    "Serialization API": [
        "GenerateSerializationByIds",
    ],
    "AASX File Server API": [
        "GetAllAASXPackageIds",
        "GetAASXByPackageId",
        "PostAASXPackage",
        "PutAASXByPackageId",
        "DeleteAASXByPackageId",
    ],
    "Asset Administration Shell Registry API": [
        "GetAllAssetAdministrationShellDescriptors",
        "GetAssetAdministrationShellDescriptorById",
        "PostAssetAdministrationShellDescriptor",
        "PutAssetAdministrationShellDescriptorById",
        "DeleteAssetAdministrationShellDescriptorById",
    ],
    "Submodel Registry API": [
        "GetAllSubmodelDescriptors",
        "GetSubmodelDescriptorById",
        "PostSubmodelDescriptor",
        "PutSubmodelDescriptorById",
        "DeleteSubmodelDescriptorById",
    ],
    "Asset Administration Shell Repository API": [
        "GetAllAssetAdministrationShells",
        "GetAllAssetAdministrationShells-Reference",
        "GetAssetAdministrationShellById",
        "PostAssetAdministrationShell",
        "PutAssetAdministrationShellById",
        "DeleteAssetAdministrationShellById",
        "GetAssetAdministrationShellById-Reference_AasRepository",
    ],
    "Submodel Repository API": [
        "GetAllSubmodelDescriptors",
        "GetSubmodelDescriptorById",
        "PostSubmodelDescriptor",
        "PutSubmodelDescriptorById",
        "DeleteSubmodelDescriptorById",
    ],
    "Concept Description Repository API": [
        "GetAllConceptDescriptions",
        "GetConceptDescriptionById",
        "PostConceptDescription",
        "PutConceptDescriptionById",
        "DeleteConceptDescriptionById",
    ],
    "Asset Administration Shell Basic Discovery API": [
        "GetAllAssetAdministrationShellIdsByAssetLink",
        "GetAllAssetLinksById",
        "PostAllAssetLinksById",
        "DeleteAllAssetLinksById",
    ],
    "Description API": [
        "GetDescription",
    ],
    # Service Specs
    "Asset Administration Shell Service Specification": [
        "Asset Administration Shell API",
        "Submodel API",  # TODO: via super path
        "Serialization API",
        "Description API",
    ],
    "Submodel Service Specification": [
        "Submodel API",
        "Serialization API",
        "Description API",
    ],
    "AASX File Server Service Specification": [
        "AASX File Server API",
        "Description API",
    ],
    "Asset Administration Shell Registry Service Specification": [
        "Asset Administration Shell Registry API",
        "Submodel Registry API",  # TODO: via super path
        "Description API",
    ],
    "Submodel Registry Service Specification": [
        "Submodel Registry API",
        "Description API",
    ],
    "Discovery Service Specification": [
        "Asset Administration Shell Basic Discovery API",
        "Description API",
    ],
    "Asset Administration Shell Repository Service Specification": [
        "Asset Administration Shell API",  # TODO: via super path
        "Submodel API",  # TODO: via super path
        "Asset Administration Shell Repository API",
        "Submodel Repository API",  # TODO: via super path
        "Serialization API",
        "Description API",
    ],
    "Submodel Repository Service Specification": [
        "Submodel API",  # TODO: via super path
        "Submodel Repository API",  # TODO: via super path
        "Serialization API",
        "Description API",
    ],
    "ConceptDescription Repository Service Specification": [
        "Serialization API",
        "Description API",
        "Concept Description Repository API",
    ],
    # Service Spec Profiles
    f"{SSP_PREFIX}AssetAdministrationShellServiceSpecification/SSP-001": [
        "Asset Administration Shell Service Specification"
    ],
    f"{SSP_PREFIX}AssetAdministrationShellServiceSpecification/SSP-002": [
        # AAS
        "GetAssetAdministrationShell",
        "GetAssetAdministrationShell-Reference",
        "GetAllSubmodelReferences",
        "GetAssetInformation",
        "GetThumbnail",
        # Submodel Repo
        "GetSubmodel_AAS",
        "GetSubmodel-Metadata_AAS",
        "GetSubmodel-ValueOnly_AAS",
        "GetSubmodelMetadata-Reference_AAS",
        "GetSubmodel-Path_AAS",
        # Submodel Elements
        "GetAllSubmodelElements_AAS",
        "GetAllSubmodelElements-Metadata_AAS",
        "GetAllSubmodelElements-ValueOnly_AAS",
        "GetAllSubmodelElementsReference_AAS",
        "GetAllSubmodelElementsPath_AAS",
        # Submodel Element
        "GetSubmodelElementByPath_AAS",
        "GetSubmodelElementByPath-Metadata_AAS",
        "GetSubmodelElementByPath-ValueOnly_AAS",
        "GetSubmodelElementByPath-Reference_AAS",
        "GetSubmodelElementByPath-Path_AAS",
        "GetFileByPath_AAS",
    ],
    f"{SSP_PREFIX}SubmodelServiceSpecification/SSP-001": [
        "Submodel Service Specification",
    ],
    f"{SSP_PREFIX}SubmodelServiceSpecification/SSP-002": [
        # Submodel API
        "GetSubmodel",
        "GetSubmodel-Metadata",
        "GetSubmodel-ValueOnly",
        "GetSubmodel-Reference",
        "GetSubmodel-Path",
        "GetAllSubmodelElements",
        "GetAllSubmodelElements-Metadata",
        "GetAllSubmodelElements-ValueOnly",
        "GetAllSubmodelElements-Reference",
        "GetAllSubmodelElements-Path",
        "GetSubmodelElementByPath",
        "GetSubmodelElementByPath-Metadata",
        "GetSubmodelElementByPath-ValueOnly",
        "GetSubmodelElementByPath-Reference",
        "GetSubmodelElementByPath-Path",
        "GetFileByPath",
        # Serialization API
        "GenerateSerializationByIds",
        # Description API
        "GetDescription",
    ],
    f"{SSP_PREFIX}SubmodelServiceSpecification/SSP-003": [
        "GetSubmodel",
        "InvokeOperation_SubmodelRepo",
        "GetDescription",
    ],
    f"{SSP_PREFIX}AasxFileServerServiceSpecification/SSP-001": [
        "AASX File Server Service Specification"
    ],
    f"{SSP_PREFIX}AssetAdministrationShellRegistryServiceSpecification/SSP-001": [
        "Asset Administration Shell Registry Service Specification",
    ],
    f"{SSP_PREFIX}AssetAdministrationShellRegistryServiceSpecification/SSP-002": [
        "GetAllAssetAdministrationShellDescriptors",
        "GetAssetAdministrationShellDescriptorById",
        "GetAllSubmodelDescriptorsThroughSuperpath",
        "GetSubmodelDescriptorByIdThroughSuperpath",
    ],
    f"{SSP_PREFIX}SubmodelRegistryServiceSpecification/SSP-001": [
        "Submodel Registry Service Specification",
    ],
    f"{SSP_PREFIX}SubmodelRegistryServiceSpecification/SSP-002": [
        "GetAllSubmodelDescriptors",
        "GetSubmodelDescriptorById",
        "GetDescription",
    ],
    f"{SSP_PREFIX}DiscoveryServiceSpecification/SSP-001": [
        "Discovery Service Specification",
    ],
    f"{SSP_PREFIX}AssetAdministrationShellRepositoryServiceSpecification/SSP-001": [
        "Asset Administration Shell Repository Service Specification",
    ],
    f"{SSP_PREFIX}AssetAdministrationShellRepositoryServiceSpecification/SSP-002": [
        # AAS Repository API:
        "GetAllAssetAdministrationShells",
        "GetAllAssetAdministrationShells-Reference",
        "GetAssetAdministrationShellById",
        "GetAssetAdministrationShellById-Reference_AasRepository",
        # AAS API by superpath:
        "GetAllSubmodelReferences_AasRepository",
        "GetAssetInformation_AasRepository",
        "GetThumbnail_AasRepository",
        # Submodel Repository API by superpath:
        "GetSubmodelById_AasRepository",
        "GetSubmodelById-Metadata_AasRepository",
        "GetSubmodelById-ValueOnly_AasRepository",
        "GetSubmodelById-Reference_AasRepository",
        "GetSubmodelById-Path_AasRepository",
        # Submodel API by superpath:
        "GetAllSubmodelElements_AasRepository",
        "GetAllSubmodelElements-Metadata_AasRepository",
        "GetAllSubmodelElements-ValueOnly_AasRepository",
        "GetAllSubmodelElements-Reference_AasRepository",
        "GetAllSubmodelElements-Path_AasRepository",
        "GetSubmodelElementByPath_AasRepository",
        "GetSubmodelElementByPath-Metadata_AasRepository",
        "GetSubmodelElementByPath-ValueOnly_AasRepository",
        "GetSubmodelElementByPath-Reference_AasRepository",
        "GetSubmodelElementByPath-Path_AasRepository",
        "GetFileByPath_AasRepository",
        # Serialization API
        "GenerateSerializationByIds",
        # Description API
        "GetDescription",
    ],
    f"{SSP_PREFIX}SubmodelRepositoryServiceSpecification/SSP-001": [
        "Submodel Repository Service Specification",
    ],
    f"{SSP_PREFIX}SubmodelRepositoryServiceSpecification/SSP-002": [
        # Submodel Repository API
        "GetAllSubmodels",
        'GetAllSubmodels-Metadata',
        'GetAllSubmodels-ValueOnly',
        'GetAllSubmodels-Reference',
        'GetAllSubmodels-Path',
        "GetSubmodelById",
        "GetSubmodelById-Metadata",
        "GetSubmodelById-ValueOnly",
        "GetSubmodelById-Reference",
        "GetSubmodelById-Path",
        # Submodel API
        'GetAllSubmodelElements_SubmodelRepository',
        'GetAllSubmodelElements-Metadata_SubmodelRepository',
        'GetAllSubmodelElements-ValueOnly_SubmodelRepo',
        'GetAllSubmodelElements-Reference_SubmodelRepo',
        'GetAllSubmodelElements-Path_SubmodelRepo',
        "GetSubmodelElementByPath_SubmodelRepo",
        "GetSubmodelElementByPath-Metadata_SubmodelRepo",
        "GetSubmodelElementByPath-ValueOnly_SubmodelRepo",
        "GetSubmodelElementByPath-Reference_SubmodelRepo",
        "GetSubmodelElementByPath-Path_SubmodelRepo",
        "GetFileByPath_SubmodelRepo",
        # Serialization API
        "GenerateSerializationByIds",
        # Description API
        "GetDescription"
    ],
    f"{SSP_PREFIX}SubmodelRepositoryServiceSpecification/SSP-003": [
        f"{SSP_PREFIX}SubmodelRepositoryServiceSpecification/SSP-001"  # TODO: Constraint AASa-003
    ],
    f"{SSP_PREFIX}SubmodelRepositoryServiceSpecification/SSP-004": [
        f"{SSP_PREFIX}SubmodelRepositoryServiceSpecification/SSP-002"  # TODO: Constraint AASa-004
    ],
    f"{SSP_PREFIX}ConceptDescriptionRepositoryServiceSpecification/SSP-001": [
        "ConceptDescription Repository Service Specification"
    ]
})

_test_suites: Dict[str, "ApiTestSuite"] = {}


def operation(name: str):
    assert name not in _test_suites

    def decorator(fn):
        _test_suites[name] = fn
        return fn
    return decorator


@dataclass
class ExecConf:
    server: str
    dry: bool = False
    verify: bool = True
    remove_path_prefix: str = ""


def _check_server(exec_conf: ExecConf) -> bool:
    with start(f'Trying to reach {exec_conf.server}'):
        if exec_conf.dry:
            write(AasTestResult("Skipped due to dry run", Level.WARNING))
            return True

        try:
            requests.get(exec_conf.server, verify=exec_conf.verify)
            write('OK')
            return True
        except requests.exceptions.RequestException as e:
            write(AasTestResult('Failed to reach: {}'.format(e), level=Level.CRITICAL))
            return False


class AasSpec:

    def __init__(self, open_api: OpenApi) -> None:
        self.open_api = open_api


def _load_spec() -> AasSpec:
    script_dir = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(script_dir, 'api.yml'), 'rb') as f:
        spec = load(f, Loader=Loader)
    open_api = OpenApi.from_dict(spec)
    return AasSpec(open_api)


_spec = _load_spec()


def _shorten(content: bytes, max_len: int = 300) -> str:
    try:
        content = content.decode()
    except UnicodeDecodeError:
        return "<binary-data>"
    if len(content) > max_len:
        return content[:max_len] + "..."
    return content


def _get_json(response: requests.models.Response) -> dict:
    try:
        data = response.json()
        if not isinstance(data, (dict, list)):
            abort(f"Expected JSON, got {type(data)}")
        return data
    except requests.exceptions.JSONDecodeError as e:
        abort(AasTestResult(f"Cannot decode as JSON: {e}", Level.CRITICAL))


def _invoke(request: Request, conf: ExecConf, positive_test) -> requests.models.Response:
    prepared_request = request.build(conf.server).prepare()
    response = requests.Session().send(prepared_request, verify=conf.verify)
    write(f"Response: ({response.status_code}): {_shorten(response.content)}")
    if response.status_code >= 500:
        abort(AasTestResult(f"Got status code {response.status_code}", Level.CRITICAL))
    if positive_test:
        if response.status_code < 200 or response.status_code > 299:
            abort(f"Expected status code 2xx, but got {response.status_code}")
    else:
        if response.status_code < 400 or response.status_code > 499:
            abort(f"Expected status code 4xx, but got {response.status_code}")
    return response


def _invoke_and_decode(request: Request, conf: ExecConf, positive_test: bool) -> dict:
    with start(f"Invoke: {request.operation.method.upper()} {request.make_path()}"):
        response = _invoke(request, conf, positive_test)
        expected_responses = []
        expected_responses += [i for i in request.operation.responses if i.code == response.status_code]
        expected_responses += [i for i in request.operation.responses if i.code is None]
        if not expected_responses:
            abort(f"Invalid status code {response.status_code}")
        data = _get_json(response)
        ref = expected_responses[0].schema.get('$ref')
        if ref == '#/components/schemas/AssetAdministrationShell':
            result, aas = parse_and_check_json(AssetAdministrationShell, data)
            write(result)
            return data
        if ref == '#/components/schemas/Environment':
            result, aas = parse_and_check_json(Environment, data)
            write(result)
            return data
        if ref == '#/components/schemas/GetAssetAdministrationShellsResult':
            entries = data.get('result', None)
            if entries:
                result, aas = parse_and_check_json(List[AssetAdministrationShell], entries)
                write(result)
            # Fallthrough to check against schema for paging_metadata

        validator = parse_schema({**expected_responses[0].schema, '$schema': 'https://json-schema.org/draft/2020-12/schema'}, ParseConfig(raise_on_unknown_format=False))
        validation_result = validator.validate(data)
        if validation_result.ok:
            write("Response conforms to schema")
        else:
            result = AasTestResult(f"Invalid response for schema", level=Level.ERROR)
            _map_error(result, validation_result)
            raise ResultException(result)
        return data


class ApiTestSuite:

    def __init__(self, operation: Operation, conf: ExecConf, sample_cache: SampleCache, open_api: OpenApi, suite: str):
        self.operation = operation
        self.conf = conf
        self.sample_cache = sample_cache
        self.open_api = open_api
        self.suite = suite
        self.valid_values: Dict[str, List[any]] = {}

    def setup(self):
        pass

    def execute_syntactic_tests(self):
        valid_values = {k: [v] for k, v in self.valid_values.items()}
        graph = generate_all(self.operation, self.sample_cache, valid_values)
        for i in graph.generate_paths():
            request: Request = graph.execute(i.path)
            if not i.is_valid:
                _invoke_and_decode(request, self.conf, False)

    def execute_semantic_tests(self):
        fns = [getattr(self, i) for i in dir(self) if i.startswith('test_')]
        fns.sort(key=lambda x: x.__code__.co_firstlineno)
        for test_fn in fns:
            with start(test_fn.__doc__ or test_fn.__name__):
                test_fn()

    def execute(self) -> ConfusionMatrix:
        mat = ConfusionMatrix()
        with start("Negative Tests") as result:
            self.execute_syntactic_tests()
        mat.invalid_rejected = len([i for i in result.sub_results if i.ok()])
        mat.invalid_accepted = len(result.sub_results) - mat.invalid_rejected
        with start("Positive Tests") as result:
            self.execute_semantic_tests()
        mat.valid_accepted = len([i for i in result.sub_results if i.ok()])
        mat.valid_rejected = len(result.sub_results) - mat.valid_accepted
        return mat

    def teardown(self):
        pass


# /shells
class GetAllAasTestSuiteBase(ApiTestSuite):

    def setup(self):
        self.cursor: Optional[str] = None
        op = self.open_api.operations["GetAllAssetAdministrationShells"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        self.valid_id_short: str = _lookup(data, ['result', 0, 'idShort'])
        global_asset_id = _lookup(data, ['result', 0, 'assetInformation', 'globalAssetId'], None)
        if global_asset_id:
            self.asset_id = {'name': 'globalAssetId', 'value': 'globalAssetId'}
        else:
            self.asset_id = _lookup(data, ['result', 0, 'specificAssetIds', 0])
        self.cursor = _lookup(data, ['paging_metadata', 'cursor'], None)

    def test_no_parameters(self):
        """
        Invoke without parameters
        """
        request = generate_one_valid(self.operation, self.sample_cache)
        _invoke_and_decode(request, self.conf, True)

    def test_get_one(self):
        """
        Fetch only one
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        data = _lookup(data, ['result'])
        _assert(len(data) == 1, 'Has exactly one result entry')

    def test_filter_by_non_existing_idshort(self):
        """
        Filter by non-existing idShort
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'idShort': 'does-not-exist'})
        data = _invoke_and_decode(request, self.conf, True)
        data = _lookup(data, ['result'])
        _assert(len(data) == 0, 'Result is empty')

    def test_pagination(self):
        """
        Test pagination
        """
        if self.cursor is None:
            abort(AasTestResult("Cannot check pagination, there must be at least 2 shells", level=Level.WARNING))
        request = generate_one_valid(self.operation, self.sample_cache, {'cursor': self.cursor, 'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        data = _lookup(data, ['result'])
        _assert(len(data) == 1, 'Exactly one entry')

    def test_filter_by_asset_id(self):
        """
        Filter by assetId
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'assetIds': self.asset_id})
        _invoke_and_decode(request, self.conf, True)


@operation("GetAllAssetAdministrationShells")
class GetAllAasTestSuite(GetAllAasTestSuiteBase):
    def test_filter_by_idshort(self):
        """
        Filter by idShort
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'idShort': self.valid_id_short})
        data = _invoke_and_decode(request, self.conf, True)
        id_short = _lookup(data, ['result', 0, 'idShort'])
        _assert(id_short == self.valid_id_short, 'Result has the requested idShort')


@operation("GetAllAssetAdministrationShells-Reference")
class GetAllAasRefsTestSuite(GetAllAasTestSuiteBase):
    def test_filter_by_idshort(self):
        """
        Filter by idShort
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'idShort': self.valid_id_short})
        _invoke_and_decode(request, self.conf, True)

# /shells/<AAS>


@operation("GetAssetAdministrationShellById")
class GetAasById(ApiTestSuite):
    def setup(self):
        op = self.open_api.operations["GetAllAssetAdministrationShells"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        self.valid_id: str = _lookup(data, ['result', 0, 'id'])

    def test_get(self):
        """
        Fetch AAS by id
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'aasIdentifier': b64urlsafe(self.valid_id)})
        data = _invoke_and_decode(request, self.conf, True)
        data = _lookup(data, ['id'])
        _assert(data == self.valid_id, 'Returned the correct one')


@operation('GetAssetAdministrationShellById-Reference_AasRepository')
class GetAasReferenceById(ApiTestSuite):
    def setup(self):
        op = self.open_api.operations["GetAllAssetAdministrationShells"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        self.valid_id: str = _lookup(data, ['result', 0, 'id'])

    def test_simple(self):
        """
        Fetch AAS reference by id
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'aasIdentifier': b64urlsafe(self.valid_id)})
        data = _invoke_and_decode(request, self.conf, True)
        data = _lookup(data, ['keys', 0, 'value'])
        _assert(data == self.valid_id, 'Returned the correct one')

# /shells/<AAS>/submodels


@operation("GetAllSubmodels_AasRepository")
@operation("GetAllSubmodels_AasRepository Metadata")
@operation("GetAllSubmodels_AasRepository-ValueOnly")
@operation("GetAllSubmodels_AasRepository-Reference")
@operation("GetAllSubmodels_AasRepository-Path")
class GetAllSubmodelsTestSuite(ApiTestSuite):
    def setup(self):
        op = self.open_api.operations["GetAllAssetAdministrationShells"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        self.valid_id = _lookup(data, ['result', 0, 'id'])

    def test_simple(self):
        """
        Fetch by id
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'aasIdentifier': b64urlsafe(self.valid_id)})
        _invoke_and_decode(request, self.conf, True)


# /shells/<AAS>/submodel-refs


@operation("GetAllSubmodelReferences_AasRepository")
class GetAllSubmodelRefsTestSuite(ApiTestSuite):
    def setup(self):
        op = self.open_api.operations["GetAllAssetAdministrationShells"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        self.valid_id = _lookup(data, ['result', 0, 'id'])
        request = generate_one_valid(self.operation, self.sample_cache, {'aasIdentifier': b64urlsafe(self.valid_id), 'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        self.cursor = _lookup(data, ['paging_metadata', 'cursor'], None)

    def test_simple(self):
        """
        Fetch by id
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'aasIdentifier': b64urlsafe(self.valid_id)})
        _invoke_and_decode(request, self.conf, True)

    def test_pagination(self):
        """
        Check pagination
        """
        if self.cursor is None:
            abort(AasTestResult("Cannot check pagination, there must be at least 2 submodels", level=Level.WARNING))
        request = generate_one_valid(self.operation, self.sample_cache, {'aasIdentifier': b64urlsafe(self.valid_id), 'cursor': self.cursor, 'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        data = _lookup(data, ['result'])
        _assert(len(data) == 1, 'Exactly one entry')


# /shells/<AAS>/asset-information


class AssetInfoTestSuiteBase(ApiTestSuite):
    def setup(self):
        op = self.open_api.operations["GetAllAssetAdministrationShells"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        self.valid_id = _lookup(data, ['result', 0, 'id'])
        self.valid_values = {
            'aasIdentifier': [b64urlsafe(self.valid_id)]
        }


@operation("GetAssetInformation_AasRepository")
class AssetInfoBySuperPathSuite(AssetInfoTestSuiteBase):
    def test_simple(self):
        """
        Fetch by id
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'aasIdentifier': b64urlsafe(self.valid_id)})
        _invoke_and_decode(request, self.conf, True)


@operation('GetThumbnail_AasRepository')
class AasThumbnailBySuperPathSuite(AssetInfoTestSuiteBase):
    def setup(self):
        op = self.open_api.operations["GetAllAssetAdministrationShells"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        self.valid_id: str = _lookup(data, ['result', 0, 'id'])

    def test_simple(self):
        """
        Fetch thumbnail by id
        """
        request = generate_one_valid(self.operation, self.sample_cache, {
            'aasIdentifier': b64urlsafe(self.valid_id),
        })
        _invoke(request, self.conf, True)

###########################
#     Submodel Interface
##########################


class SubmodelByAasRepoSuite(ApiTestSuite):
    def setup(self):
        op = self.open_api.operations["GetAllAssetAdministrationShells"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        valid_id: str = _lookup(data, ['result', 0, 'id'])
        self.valid_submodel_id: str = _lookup(data, ['result', 0, 'submodels', 0, 'keys', 0, 'value'])
        self.valid_values = {
            'aasIdentifier': b64urlsafe(valid_id),
            'submodelIdentifier': b64urlsafe(self.valid_submodel_id),
        }


class SubmodelBySubmodelRepoSuite(ApiTestSuite):
    def setup(self):
        op = self.open_api.operations["GetAllSubmodels"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        self.valid_submodel_id: str = _lookup(data, ['result', 0, 'id'])
        self.valid_values = {
            'submodelIdentifier': b64urlsafe(self.valid_submodel_id),
        }


class SubmodelSuite(ApiTestSuite):
    def setup(self):
        self.valid_submodel_id = None
        self.valid_values = {}

# /submodels


@operation("GetAllSubmodels")
@operation("GetAllSubmodels-Path")
@operation("GetAllSubmodels-Reference")
@operation("GetAllSubmodels-Metadata")
@operation("GetAllSubmodels-ValueOnly")
class GetAllSubmodelsTests(ApiTestSuite):
    def test_simple(self):
        """
        Fetch all submodels
        """
        request = generate_one_valid(self.operation, self.sample_cache, self.valid_values)
        _invoke_and_decode(request, self.conf, True)


# /shells/<AAS>/submodels/<SM>
# /submodels/<SM>
# /submodel


class GetSubmodelTests(ApiTestSuite):
    def test_simple(self):
        """
        Fetch submodel by id
        """
        request = generate_one_valid(self.operation, self.sample_cache, self.valid_values)
        data = _invoke_and_decode(request, self.conf, True)
        data = _lookup(data, ['id'])
        if self.valid_submodel_id:
            _assert(data == self.valid_submodel_id, 'Returns the correct one')


@operation("GetSubmodelById_AasRepository")
@operation("GetSubmodelById-Metadata_AasRepository")
class GetSubmodelById_AasRepository(SubmodelByAasRepoSuite, GetSubmodelTests):
    pass


@operation("GetSubmodelById")
@operation("GetSubmodelById-Metadata")
class GetSubmodelById(SubmodelBySubmodelRepoSuite, GetSubmodelTests):
    pass


@operation("GetSubmodel")
@operation("GetSubmodel-Metadata")
class GetSubmodel(SubmodelSuite, GetSubmodelTests):
    pass


class GetSubmodelRefsTests(ApiTestSuite):
    def test_simple(self):
        """
        Fetch submodel by id
        """
        request = generate_one_valid(self.operation, self.sample_cache, self.valid_values)
        _invoke_and_decode(request, self.conf, True)


@operation("GetSubmodelById-ValueOnly_AasRepository")
@operation("GetSubmodelById-Reference_AasRepository")
@operation("GetSubmodelById-Path_AasRepository")
class GetSubmodelById_Reference_AasRepository(SubmodelByAasRepoSuite, GetSubmodelRefsTests):
    pass


@operation("GetSubmodelById-ValueOnly")
@operation("GetSubmodelById-Reference")
@operation("GetSubmodelById-Path")
class GetSubmodelById_Reference(SubmodelBySubmodelRepoSuite, GetSubmodelRefsTests):
    pass


@operation("GetSubmodel-ValueOnly")
@operation("GetSubmodel-Reference")
@operation("GetSubmodel-Path")
class GetSubmodel_Reference(SubmodelSuite, GetSubmodelRefsTests):
    pass

# /shells/<AAS>/submodels/<SM>/submodel-elements
# /shells/<AAS>/submodels/<SM>/submodel-elements


class GetSubmodelElementTests(ApiTestSuite):
    def test_simple(self):
        """
        Fetch all submodel elements
        """
        request = generate_one_valid(self.operation, self.sample_cache, self.valid_values)
        _invoke_and_decode(request, self.conf, True)

    def test_level_core(self):
        """
        Fetch all submodel elements with level=core
        """
        request = generate_one_valid(self.operation, self.sample_cache, {
            **self.valid_values,
            'level': 'core'
        })
        _invoke_and_decode(request, self.conf, True)

    def test_level_deep(self):
        """
        Fetch all submodel elements with level=deep
        """
        request = generate_one_valid(self.operation, self.sample_cache, {
            **self.valid_values,
            'level': 'deep'
        })
        _invoke_and_decode(request, self.conf, True)

    def test_extent_with_blob_value(self):
        """
        Fetch all submodel elements with extent=withBlobValue
        """
        request = generate_one_valid(self.operation, self.sample_cache, {
            **self.valid_values,
            'extent': 'withBlobValue'
        })
        _invoke_and_decode(request, self.conf, True)

    def test_extent_without_blob_value(self):
        """
        Fetch all submodel elements with extent=withoutBlobValue
        """
        request = generate_one_valid(self.operation, self.sample_cache, {
            **self.valid_values,
            'extent': 'withoutBlobValue'
        })
        _invoke_and_decode(request, self.conf, True)


@operation("GetAllSubmodelElements_AasRepository")
class GetAllSubmodelElements_AasRepository(SubmodelByAasRepoSuite, GetSubmodelElementTests):
    pass


@operation("GetAllSubmodelElements_SubmodelRepository")
class GetAllSubmodelElements_SubmodelRepository(SubmodelBySubmodelRepoSuite, GetSubmodelElementTests):
    pass


@operation("GetAllSubmodelElements")
class GetAllSubmodelElements(SubmodelSuite, GetSubmodelElementTests):
    pass


class GetAllSubmodelElementsValueOnlyTests(ApiTestSuite):
    def test_simple(self):
        """
        Fetch all submodel elements
        """
        request = generate_one_valid(self.operation, self.sample_cache, self.valid_values)
        _invoke_and_decode(request, self.conf, True)

    def test_level_core(self):
        """
        Fetch all submodel elements with level=core
        """
        request = generate_one_valid(self.operation, self.sample_cache, {
            **self.valid_values,
            'level': 'core'
        })
        _invoke_and_decode(request, self.conf, True)

    def test_level_deep(self):
        """
        Fetch all submodel elements with level=deep
        """
        request = generate_one_valid(self.operation, self.sample_cache, {
            **self.valid_values,
            'level': 'deep'
        })
        _invoke_and_decode(request, self.conf, True)


@operation("GetAllSubmodelElements-ValueOnly_AasRepository")
@operation("GetAllSubmodelElements-Reference_AasRepository")
@operation("GetAllSubmodelElements-Path_AasRepository")
class GetAllSubmodelElements_ValueOnly_AasRepository(SubmodelByAasRepoSuite, GetAllSubmodelElementsValueOnlyTests):
    pass


@operation("GetAllSubmodelElements-ValueOnly_SubmodelRepo")
@operation("GetAllSubmodelElements-Reference_SubmodelRepo")
@operation("GetAllSubmodelElements-Path_SubmodelRepo")
class GetAllSubmodelElements_ValueOnly_SubmodelRepository(SubmodelBySubmodelRepoSuite, GetAllSubmodelElementsValueOnlyTests):
    pass


@operation("GetAllSubmodelElements-ValueOnly")
@operation("GetAllSubmodelElements-Reference")
@operation("GetAllSubmodelElements-Path")
class GetAllSubmodelElements_ValueOnly(SubmodelSuite, GetAllSubmodelElementsValueOnlyTests):
    pass


class GetAllSubmodelElementRefsTests(ApiTestSuite):
    def test_simple(self):
        """
        Fetch all submodel elements
        """
        request = generate_one_valid(self.operation, self.sample_cache, self.valid_values)
        _invoke_and_decode(request, self.conf, True)


@operation("GetAllSubmodelElements-Metadata_AasRepository")
class GetAllSubmodelElements_Metadata_AasRepository(SubmodelByAasRepoSuite, GetAllSubmodelElementRefsTests):
    pass


@operation("GetAllSubmodelElements-Metadata_SubmodelRepository")
class GetAllSubmodelElements_Metadata_SubmodelRepository(SubmodelBySubmodelRepoSuite, GetAllSubmodelElementRefsTests):
    pass


@operation("GetAllSubmodelElements-Metadata")
class GetAllSubmodelElements_Metadata(SubmodelSuite, GetAllSubmodelElementRefsTests):
    pass


# /shells/<AAS>/submodels/<SM>/submodel-elements/<ID>

class SubmodelElementTestsBase(ApiTestSuite):
    all_submodel_elements = [
        'SubmodelElementCollection',
        'SubmodelElementList',
        'Entity',
        'BasicEventElement',
        'Capability',
        'Operation',
        'Property',
        'MultiLanguageProperty',
        'Range',
        'ReferenceElement',
        'RelationshipElement',
        'AnnotatedRelationshipElement',
        'Blob',
        'File',
    ]


def _collect_submodel_elements(data: list, paths: Dict[str, List[str]], path_prefix: str):
    for i in data:
        id_short = path_prefix + _lookup(i, ['idShort'])
        model_type = _lookup(i, ['modelType'])
        try:
            paths[model_type].append(id_short)
        except KeyError:
            paths[model_type] = [id_short]
        if model_type == 'SubmodelElementCollection':
            value = _lookup(i, ['value'])
            _collect_submodel_elements(value, paths, id_short + ".")


class SubmodelElementByAasRepoSuite(SubmodelElementTestsBase):

    def setup(self):
        self.paths = {}
        op = self.open_api.operations["GetAllAssetAdministrationShells"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        valid_id = _lookup(data, ['result', 0, 'id'])
        valid_submodel_id = _lookup(data, ['result', 0, 'submodels', 0, 'keys', 0, 'value'])
        self.valid_values = {
            'aasIdentifier': b64urlsafe(valid_id),
            'submodelIdentifier': b64urlsafe(valid_submodel_id),
        }
        op = self.open_api.operations["GetAllSubmodelElements_AasRepository"]
        request = generate_one_valid(op, self.sample_cache, self.valid_values)
        data = _invoke_and_decode(request, self.conf, True)
        elements = _lookup(data, ['result'])
        _collect_submodel_elements(elements, self.paths, '')
        self.valid_values['idShortPath'] = _first_iterator_item(self.paths.values())[0]


class SubmodelElementBySubmodelRepoSuite(SubmodelElementTestsBase):

    def setup(self):
        self.paths = {}
        op = self.open_api.operations["GetAllSubmodels"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        valid_submodel_id = _lookup(data, ['result', 0, 'id'])
        self.valid_values = {
            'submodelIdentifier': b64urlsafe(valid_submodel_id),
        }
        op = self.open_api.operations["GetAllSubmodelElements_SubmodelRepository"]
        request = generate_one_valid(op, self.sample_cache, self.valid_values)
        data = _invoke_and_decode(request, self.conf, True)
        elements = _lookup(data, ['result'])
        _collect_submodel_elements(elements, self.paths, '')
        self.valid_values['idShortPath'] = _first_iterator_item(self.paths.values())[0]


class SubmodelElementBySubmodelSuite(SubmodelElementTestsBase):
    def setup(self):
        self.valid_values = {}


class SubmodelElementBySuperpathSuite(SubmodelElementTestsBase):
    supported_submodel_elements = {
        'SubmodelElementCollection',
        'SubmodelElementList',
        'Entity',
        'BasicEventElement',
        'Capability',
        'Operation',
        'Property',
        'MultiLanguageProperty',
        'Range',
        'ReferenceElement',
        'RelationshipElement',
        'AnnotatedRelationshipElement',
        'Blob',
        'File',
    }

    def check_type(self, model_type: str, level: Optional[str], extent: Optional[str]):
        if model_type not in self.paths:
            abort(AasTestResult("No such element present", level=Level.WARNING))
        id_short_path = self.paths[model_type][0]
        valid_values = self.valid_values.copy()
        valid_values['idShortPath'] = [id_short_path]
        if level:
            valid_values['level'] = level
        if extent:
            valid_values['extent'] = extent

        request = generate_one_valid(self.operation, self.sample_cache, valid_values)
        data = _invoke_and_decode(request, self.conf, model_type in self.supported_submodel_elements)

    def test_no_params(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level=None, extent='')

    def test_level_deep(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level='deep', extent='')

    def test_level_core(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level='core', extent='')

    def test_extend_with_blob_value(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level=None, extent='withBlobValue')

    def test_extend_without_blob_value(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level=None, extent='withoutBlobValue')


@operation("GetSubmodelElementByPath_AasRepository")
class GetSubmodelElementByPath_AasRepository(SubmodelElementByAasRepoSuite, GetSubmodelElementTests):
    pass


@operation("GetSubmodelElementByPath_SubmodelRepo")
class GetSubmodelElementByPath_SubmodelRepository(SubmodelElementBySubmodelRepoSuite, GetSubmodelElementTests):
    pass


@operation("GetSubmodelElementByPath")
class GetSubmodelElementByPath(SubmodelElementBySubmodelSuite, GetSubmodelElementTests):
    pass


class GetSubmodelElementValueOnlyTests(SubmodelElementTestsBase):
    supported_submodel_elements = {
        'SubmodelElementCollection',
        'SubmodelElementList',
        'Entity',
        'BasicEventElement',
        'Property',
        'MultiLanguageProperty',
        'Range',
        'ReferenceElement',
        'RelationshipElement',
        'AnnotatedRelationshipElement',
        'Blob',
        'File',
    }

    def check_type(self, model_type: str, level: Optional[str], extent: Optional[str]):
        if model_type not in self.paths:
            abort(AasTestResult("No such element present", level=Level.WARNING))
        id_short_path = self.paths[model_type][0]
        valid_values = self.valid_values.copy()
        valid_values['idShortPath'] = [id_short_path]
        if level:
            valid_values['level'] = level
        if extent:
            valid_values['extent'] = extent

        request = generate_one_valid(self.operation, self.sample_cache, valid_values)
        data = _invoke_and_decode(request, self.conf, model_type in self.supported_submodel_elements)

    def test_no_params(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level=None, extent=None)

    def test_level_deep(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level='deep', extent=None)

    def test_level_core(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level='core', extent=None)

    def test_extend_with_blob_value(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level=None, extent='withBlobValue')

    def test_extend_without_blob_value(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level=None, extent='withoutBlobValue')


@operation("GetSubmodelElementByPath-ValueOnly_AasRepository")
class GetSubmodelElementByPath_ValueOnly_AasRepository(SubmodelElementByAasRepoSuite, GetSubmodelElementValueOnlyTests):
    pass


@operation("GetSubmodelElementByPath-ValueOnly_SubmodelRepo")
class GetSubmodelElementByPath_ValueOnly_SubmodelRepository(SubmodelElementBySubmodelRepoSuite, GetSubmodelElementValueOnlyTests):
    pass


class GetSubmodelElementPathTests(SubmodelElementTestsBase):
    supported_submodel_elements = {
        'SubmodelElementCollection',
        'SubmodelElementList',
        'Entity',
    }

    def check_type(self, model_type: str, level: Optional[str]):
        if model_type not in self.paths:
            abort(AasTestResult("No such element present", level=Level.WARNING))
        id_short_path = self.paths[model_type][0]
        valid_values = self.valid_values.copy()
        valid_values['idShortPath'] = [id_short_path]
        if level:
            valid_values['level'] = level

        request = generate_one_valid(self.operation, self.sample_cache, valid_values)
        data = _invoke_and_decode(request, self.conf, model_type in self.supported_submodel_elements)

    def test_no_params(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level=None)

    def test_level_deep(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level='deep')

    def test_level_core(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type, level='core')


@operation("GetSubmodelElementByPath-Path_AasRepository")
class GetSubmodelElementByPath_Path_AasRepository(GetSubmodelElementPathTests, SubmodelElementByAasRepoSuite):
    pass


@operation("GetSubmodelElementByPath-Path_SubmodelRepo")
class GetSubmodelElementByPath_Path_SubmodelRepo(GetSubmodelElementPathTests, SubmodelElementBySubmodelRepoSuite):
    pass


class GetSubmodelElementReferenceTests(SubmodelElementTestsBase):
    supported_submodel_elements = {
        'SubmodelElementCollection',
        'SubmodelElementList',
        'Entity',
        'BasicEventElement',
        'Capability',
        'Operation',
        'Property',
        'MultiLanguageProperty',
        'Range',
        'ReferenceElement',
        'RelationshipElement',
        'AnnotatedRelationshipElement',
        'Blob',
        'File',
    }

    def check_type(self, model_type: str):
        if model_type not in self.paths:
            abort(AasTestResult("No such element present", level=Level.WARNING))
        id_short_path = self.paths[model_type][0]
        valid_values = self.valid_values.copy()
        valid_values['idShortPath'] = [id_short_path]
        request = generate_one_valid(self.operation, self.sample_cache, valid_values)
        data = _invoke_and_decode(request, self.conf, model_type in self.supported_submodel_elements)

    def test_no_params(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type)


@operation("GetSubmodelElementByPath-Reference_AasRepository")
class GetSubmodelElementByPath_Path_AasRepository(GetSubmodelElementReferenceTests, SubmodelElementByAasRepoSuite):
    pass


@operation("GetSubmodelElementByPath-Reference_SubmodelRepo")
class GetSubmodelElementByPath_Path_SubmodelRepository(GetSubmodelElementReferenceTests, SubmodelElementBySubmodelRepoSuite):
    pass


class GetSubmodelElementMetadataTests(SubmodelElementTestsBase):
    supported_submodel_elements = {
        'SubmodelElementCollection',
        'SubmodelElementList',
        'Entity',
        'BasicEventElement',
        'Property',
        'MultiLanguageProperty',
        'Range',
        'ReferenceElement',
        'RelationshipElement',
        'AnnotatedRelationshipElement',
        'Blob',
        'File',
    }

    def check_type(self, model_type: str):
        if model_type not in self.paths:
            abort(AasTestResult("No such element present", level=Level.WARNING))
        id_short_path = self.paths[model_type][0]
        valid_values = self.valid_values.copy()
        valid_values['idShortPath'] = [id_short_path]
        request = generate_one_valid(self.operation, self.sample_cache, valid_values)
        data = _invoke_and_decode(request, self.conf, model_type in self.supported_submodel_elements)

    def test_no_params(self):
        for model_type in self.all_submodel_elements:
            with start(f"Checking {model_type}"):
                self.check_type(model_type)


@operation("GetSubmodelElementByPath-Metadata_AasRepository")
class GetSubmodelElementByPath_Metadata_AasRepository(GetSubmodelElementMetadataTests, SubmodelElementByAasRepoSuite):
    pass


@operation("GetSubmodelElementByPath-Metadata_SubmodelRepo")
class GetSubmodelElementByPath_Metadata_SubmodelRepository(GetSubmodelElementMetadataTests, SubmodelElementBySubmodelRepoSuite):
    pass


class GetFileByPathTests(SubmodelElementTestsBase):

    def test_no_params(self):
        """
        Invoke without params
        """
        try:
            self.valid_values['idShortPath'] = self.paths['File']
        except KeyError:
            abort("No submodel element of type 'File' found, skipping test.")
        request = generate_one_valid(self.operation, self.sample_cache, self.valid_values)
        _invoke(request, self.conf, True)


@operation("GetFileByPath_AasRepository")
class GetSubmodelElementByPath_Metadata_AasRepository(GetFileByPathTests, SubmodelElementByAasRepoSuite):
    pass


@operation("GetFileByPath_SubmodelRepo")
class GetSubmodelElementByPath_Metadata_SubmodelRepository(GetFileByPathTests, SubmodelElementBySubmodelRepoSuite):
    pass

# /serialization


@operation("GenerateSerializationByIds")
class GenerateSerializationSuite(ApiTestSuite):
    def setup(self):
        op = self.open_api.operations["GetAllAssetAdministrationShells"]
        request = generate_one_valid(op, self.sample_cache, {'limit': 1})
        data = _invoke_and_decode(request, self.conf, True)
        valid_id = _lookup(data, ['result', 0, 'id'])
        valid_submodel_id = _lookup(data, ['result', 0, 'submodels', 0, 'keys', 0, 'value'])
        self.valid_aas_id = b64urlsafe(valid_id)
        self.valid_submod_id = b64urlsafe(valid_submodel_id)

    def test_filter_by_aasids(self):
        """
        Filter by aas ids
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'aasIds': self.valid_aas_id})
        _invoke_and_decode(request, self.conf, True)

    def test_filter_by_(self):
        """
        Filter by submodel ids
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'submodelIds': self.valid_submod_id})
        _invoke_and_decode(request, self.conf, True)

    def test_include_concept_descriptions(self):
        """
        Invoke with includeConceptDescriptions
        """
        request = generate_one_valid(self.operation, self.sample_cache, {'includeConceptDescriptions': True})
        data = _invoke_and_decode(request, self.conf, True)
        _assert('conceptDescriptions' in data, 'contains conceptDescriptions', Level.WARNING)


# /description

@operation("GetDescription")
class GetDescriptionTestSuite(ApiTestSuite):
    def test_contains_suite(self):
        """
        Returned profiles must contain suite
        """
        request = generate_one_valid(self.operation, self.sample_cache)
        data = _invoke_and_decode(request, self.conf, True)
        suites = _lookup(data, ['profiles'])
        _assert(self.suite in suites, f"Contains {self.suite}", Level.WARNING)


@operation('PostAssetAdministrationShell')
@operation('PutAssetAdministrationShellById')
@operation('DeleteAssetAdministrationShellById')
@operation('PutAssetInformation_AasRepository')
@operation('PutThumbnail_AasRepository')
@operation('DeleteThumbnail_AasRepository')
@operation('PostSubmodelReference_AasRepository')
@operation('DeleteSubmodelReferenceById_AasRepository')
@operation('PutSubmodelById_AasRepository')
@operation('DeleteSubmodelById_AasRepository')
@operation('PatchSubmodel_AasRepository')
@operation('PatchSubmodelById-Metadata_AasRepository')
@operation('PatchSubmodelById-ValueOnly_AasRepository')
@operation("PostSubmodelElement_AasRepository")
@operation('PutSubmodelElementByPath_AasRepository')
@operation('PostSubmodelElementByPath_AasRepository')
@operation('DeleteSubmodelElementByPath_AasRepository')
@operation('PatchSubmodelElementValueByPath_AasRepository')
@operation('PatchSubmodelElementValueByPath-Metadata')
@operation('PatchSubmodelElementValueByPath-ValueOnly')
@operation('PutFileByPath_AasRepository')
@operation('DeleteFileByPath_AasRepository')
@operation('InvokeOperation_AasRepository')
@operation('InvokeOperation-ValueOnly_AasRepository')
@operation('InvokeOperationAsync_AasRepository')
@operation('InvokeOperationAsync-ValueOnly_AasRepository')
@operation('GetOperationAsyncStatus_AasRepository')
@operation('GetOperationAsyncResult_AasRepository')
@operation('GetOperationAsyncResult-ValueOnly_AasRepository')
# /aas
@operation('GetAssetAdministrationShell')
@operation('PutAssetAdministrationShell')
@operation('GetAssetAdministrationShell-Reference')
@operation('GetAssetInformation')
@operation('PutAssetInformation')
@operation('GetThumbnail')
@operation('PutThumbnail')
@operation('DeleteThumbnail')
# /aas/submodel-refs
@operation('GetAllSubmodelReferences')
@operation('GetAllSubmodelReferences')
@operation('PostSubmodelReference')
@operation('DeleteSubmodelReferenceById')
# /aas/submodels/<SM>
@operation('GetSubmodel_AAS')
@operation('PutSubmodel_AAS')
@operation('DeleteSubmodelById_AAS')
@operation('PatchSubmodel_AAS')
@operation('GetSubmodel-Metadata_AAS')
@operation('PatchSubmodelMetadata_AAS')
@operation('GetSubmodel-ValueOnly_AAS')
@operation('PatchSubmodel-ValueOnly_AAS')
@operation('GetSubmodelMetadata-Reference_AAS')
@operation('GetSubmodel-Path_AAS')
# /aas/submodels/<SM>/submodel-elements
@operation('GetAllSubmodelElements_AAS')
@operation('PostSubmodelElement_AAS')
@operation('GetAllSubmodelElements-Metadata_AAS')
@operation('GetAllSubmodelElements-ValueOnly_AAS')
@operation('GetAllSubmodelElementsReference_AAS')
@operation('GetAllSubmodelElementsPath_AAS')
@operation('GetSubmodelElementByPath_AAS')
@operation('PutSubmodelElementByPath_AAS')
@operation('PostSubmodelElementByPath_AAS')
@operation('DeleteSubmodelElementByPath_AAS')
@operation('PatchSubmodelElementValueByPath_AAS')
@operation('GetSubmodelElementByPath-Metadata_AAS')
@operation('PatchSubmodelElementValueByPath-Metadata_AAS')
@operation('GetSubmodelElementByPath-ValueOnly_AAS')
@operation('PatchSubmodelElementValueByPathValueOnly_AAS')
@operation('GetSubmodelElementByPath-Reference_AAS')
@operation('GetSubmodelElementByPath-Path_AAS')
@operation('GetFileByPath_AAS')
@operation('PutFileByPath_AAS')
@operation('DeleteFileByPath_AAS')
@operation('InvokeOperationSync_AAS')
@operation('InvokeOperationSync-ValueOnly_AAS')
@operation('InvokeOperationAsync_AAS')
@operation('InvokeOperationAsync-ValueOnly_AAS')
@operation('GetOperationAsyncStatus_AAS')
@operation('GetOperationAsyncResult_AAS')
@operation('GetOperationAsyncResult-ValueOnly_AAS')
# /submodel
@operation('PutSubmodel')
@operation('PatchSubmodel')
@operation('PatchSubmodel-Metadata')
@operation('PatchSubmodel-ValueOnly')
# /submodel/submodel-elements
@operation('PostSubmodelElement')
@operation('PutSubmodelElementByPath')
@operation('PostSubmodelElementByPath')
@operation('DeleteSubmodelElementByPath')
@operation('PatchSubmodelElementByPath')
@operation('GetSubmodelElementByPath-Metadata')
@operation('PatchSubmodelElementByPath-Metadata')
@operation('GetSubmodelElementByPath-ValueOnly')
@operation('PatchSubmodelElementByPath-ValueOnly')
@operation('GetSubmodelElementByPath-Reference')
@operation('GetSubmodelElementByPath-Path')
@operation('GetFileByPath')
@operation('PutFileByPath')
@operation('DeleteFileByPath')
@operation('InvokeOperation')
@operation('InvokeOperationAsync')
@operation('InvokeOperationSync-ValueOnly')
@operation('InvokeOperationAsync-ValueOnly')
@operation('GetOperationAsyncStatus')
@operation('GetOperationAsyncResult')
@operation('GetOperationAsyncResult-ValueOnly')
# /submodels
@operation('PostSubmodel')
# /submodels/<SM>
@operation('PutSubmodelById')
@operation('DeleteSubmodelById')
@operation('PatchSubmodelById')
@operation('PatchSubmodelById-Metadata')
@operation('PatchSubmodelById-ValueOnly')
# /submodels/<SM>/submodel-elements
@operation('PostSubmodelElement_SubmodelRepository')
@operation('PutSubmodelElementByPath_SubmodelRepo')
@operation('PostSubmodelElementByPath_SubmodelRepo')
@operation('DeleteSubmodelElementByPath_SubmodelRepo')
@operation('PatchSubmodelElementByPath_SubmodelRepo')
@operation('PatchSubmodelElementByPath-Metadata_SubmodelRepo')
@operation('PatchSubmodelElementByPath-ValueOnly_SubmodelRepo')
@operation('PutFileByPath_SubmodelRepo')
@operation('DeleteFileByPath_SubmodelRepo')
@operation('InvokeOperation_SubmodelRepo')
@operation('InvokeOperation-ValueOnly_SubmodelRepo')
@operation('InvokeOperationAsync_SubmodelRepo')
@operation('InvokeOperationAsync-ValueOnly_SubmodelRepo')
@operation('GetOperationAsyncStatus_SubmodelRepo')
@operation('GetOperationAsyncResult_SubmodelRepo')
@operation('GetOperationAsyncResult-ValueOnly_SubmodelRepo')
@operation('GetAllConceptDescriptions')
@operation('PostConceptDescription')
@operation('GetConceptDescriptionById')
@operation('PutConceptDescriptionById')
@operation('DeleteConceptDescriptionById')
# /shell-descriptors
@operation('GetAllAssetAdministrationShellDescriptors')
@operation('PostAssetAdministrationShellDescriptor')
@operation('GetAssetAdministrationShellDescriptorById')
@operation('PutAssetAdministrationShellDescriptorById')
@operation('DeleteAssetAdministrationShellDescriptorById')
# /shell-descriptors/<AAS>/submodel-descriptors
@operation('GetAllSubmodelDescriptorsThroughSuperpath')
@operation('PostSubmodelDescriptor-ThroughSuperpath')
@operation('GetSubmodelDescriptorByIdThroughSuperpath')
@operation('PutSubmodelDescriptorByIdThroughSuperpath')
@operation('DeleteSubmodelDescriptorByIdThroughSuperpath')
# /submodel-descriptors
@operation('GetAllSubmodelDescriptors')
@operation('PostSubmodelDescriptor')
@operation('GetSubmodelDescriptorById')
@operation('PutSubmodelDescriptorById')
@operation('DeleteSubmodelDescriptorById')
# /lookup/shells
@operation('GetAllAssetAdministrationShellIdsByAssetLink')
@operation('GetAllAssetLinksById')
@operation('PostAllAssetLinksById')
@operation('DeleteAllAssetLinksById')
# /packages
@operation('GetAllAASXPackageIds')
@operation('PostAASXPackage')
@operation('GetAASXByPackageId')
@operation('PutAASXByPackageId')
@operation('DeleteAASXByPackageId')
class SimpleSemanticTestSuite(ApiTestSuite):
    def test_semantic(self):
        """
        Perform a simple semantic test
        """
        request = generate_one_valid(self.operation, self.sample_cache)
        _invoke_and_decode(request, self.conf, True)


# Used by unit test
def check_in_sync():
    for suite in _available_suites:
        operations = _available_suites[suite]
        for op in operations:
            if op not in _spec.open_api.operations:
                raise AasTestToolsException(f"Unknown operation {op}")


def execute_tests(conf: ExecConf, suite: str) -> Tuple[AasTestResult, ConfusionMatrix]:
    try:
        operation_ids = _available_suites[suite]
    except KeyError:
        all_suites = "\n".join(sorted(_available_suites.keys()))
        raise AasTestToolsException(f"Unknown suite {suite}, must be one of:\n{all_suites}")

    sample_cache = SampleCache()
    mat = ConfusionMatrix()

    with start(f"Checking compliance to {suite}") as result_root:

        # Initial connection check
        if not _check_server(conf):
            return result_root, mat

        for operation in _spec.open_api.operations.values():
            if operation.path.startswith(conf.remove_path_prefix):
                # TODO: this will be permanent so that you cannot call this function again
                operation.path = operation.path[len(conf.remove_path_prefix):]

        # Check individual operations
        for operation in _spec.open_api.operations.values():
            if operation.operation_id not in operation_ids:
                continue
            with start(f"Checking {operation.method.upper()} {operation.path} ({operation.operation_id})"):
                if conf.dry:
                    continue

                ctr = _test_suites[operation.operation_id]
                test_suite: ApiTestSuite = ctr(operation, conf, sample_cache, _spec.open_api, suite)

                with start("Setup") as result_setup:
                    test_suite.setup()

                if result_setup.ok():
                    sub_mat = test_suite.execute()
                    mat += sub_mat
        with start("Summary:"):
            write(f"Negative tests passed: {mat.invalid_rejected} / {mat.invalid_accepted + mat.invalid_rejected}")
            write(f"Positive tests passed: {mat.valid_accepted} / {mat.valid_accepted + mat.valid_rejected}")
    return result_root, mat



from typing import List, Dict, TextIO, Union, Any, Set, Optional, Generator
import os
import json
from yaml import load
try:
    # This one is faster but not available on all systems
    from yaml import CSafeLoader as Loader
except ImportError:
    from yaml import SafeLoader as Loader

from .exception import AasTestToolsException
from .result import AasTestResult, Level
from .data_types import validators
from ._generate import generate_graph

from xml.etree import ElementTree
from json_schema_tool.schema import SchemaValidator, SchemaValidator, ParseConfig, parse_schema
import zipfile

from aas_test_engines.test_cases.v3_0 import json_to_env, xml_to_env
from aas_test_engines.test_cases.v3_0.submodel_templates import supported_templates

from ._util import un_group, normpath, splitpath

JSON = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]


class AasSchema:

    def __init__(self, validator: SchemaValidator, schema: JSON):
        self.validator = validator
        self.schema = schema


def _find_schemas() -> Dict[str, AasSchema]:
    result = {}
    script_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(script_dir, 'data', 'file')
    for i in os.listdir(data_dir):
        path = os.path.join(data_dir, i)
        if not i.endswith('.yml'):
            continue
        schema = load(open(path, "rb"), Loader=Loader)
        config = ParseConfig(
            format_validators=validators
        )
        validator = parse_schema(schema, config)
        result[i[:-4]] = AasSchema(validator, schema)
    return result


_schemas = _find_schemas()
_DEFAULT_VERSION = '3.0'


def supported_versions() -> Dict[str, List[str]]:
    return {'3.0': supported_templates()}


def latest_version():
    return _DEFAULT_VERSION


def _get_schema(version: str) -> AasSchema:
    try:
        schema = _schemas[version]
    except KeyError:
        raise AasTestToolsException(f"Unknown version {version}, must be one of {supported_versions()}")
    return schema


def check_json_data(data: any, version: str = _DEFAULT_VERSION) -> AasTestResult:
    result, env = json_to_env(data)
    return result


def check_json_file(file: TextIO, version: str = _DEFAULT_VERSION) -> AasTestResult:
    try:
        data = json.load(file)
    except json.decoder.JSONDecodeError as e:
        return AasTestResult(f"Invalid JSON: {e}", Level.ERROR)
    return check_json_data(data, version)


def check_xml_data(data: ElementTree, version: str = _DEFAULT_VERSION, submodel_templates: Set[str] = set()) -> AasTestResult:
    result, env = xml_to_env(data)
    return result


def check_xml_file(file: TextIO, version: str = _DEFAULT_VERSION) -> AasTestResult:
    try:
        data = ElementTree.fromstring(file.read())
    except ElementTree.ParseError as e:
        return AasTestResult(f"Invalid xml: {e}", Level.ERROR)
    return check_xml_data(data, version)


NS_CONTENT_TYPES = "{http://schemas.openxmlformats.org/package/2006/content-types}"
NS_RELATIONSHIPS = "{http://schemas.openxmlformats.org/package/2006/relationships}"

TYPE_AASX_ORIGIN = 'http://admin-shell.io/aasx/relationships/aasx-origin'
TYPE_AASX_SPEC = 'http://admin-shell.io/aasx/relationships/aas-spec'
TYPE_AASX_SUPPL = 'http://admin-shell.io/aasx/relationships/aas-suppl'
DEPRECATED_TYPES = {
    'http://www.admin-shell.io/aasx/relationships/aasx-origin': TYPE_AASX_ORIGIN,
    'http://www.admin-shell.io/aasx/relationships/aas-spec': TYPE_AASX_SPEC,
    'http://www.admin-shell.io/aasx/relationships/aas-suppl': TYPE_AASX_SUPPL
}
TYPE_THUMBNAIL = 'http://schemas.openxmlformats.org/package/2006/relationships/metadata/thumbnail'


class Relationship:

    def __init__(self, type: str, target: str) -> None:
        self.type = type
        self.target = target
        self.sub_rels: List[Relationship] = []

    def sub_rels_by_type(self, type: str) -> List["Relationship"]:
        return [i for i in self.sub_rels if i.type == type]

    def dump(self, indent=0):
        type_suffix = self.type.split("/")[-1]
        print("  " * indent + f"{self.target} [{type_suffix}]")
        for sub_rel in self.sub_rels:
            sub_rel.dump(indent + 1)


def _check_content_type(zipfile: zipfile.ZipFile) -> AasTestResult:
    content_types_xml = '[Content_Types].xml'
    result = AasTestResult(f'Checking {content_types_xml}')
    try:
        with zipfile.open(content_types_xml, 'r') as f:
            content_types = ElementTree.parse(f)
            expected_tag = f'{NS_CONTENT_TYPES}Types'
            if content_types.getroot().tag != expected_tag:
                result.append(
                    AasTestResult(f"root must have tag {expected_tag}, got {content_types.getroot().tag}", Level.ERROR))
    except KeyError:
        result.append(AasTestResult(f"{content_types_xml} not found", Level.ERROR))

    return result


def _scan_relationships(zipfile: zipfile.ZipFile, parent_rel: Relationship, dir: str, file: str, visited_targets: Set[str]) -> Optional[AasTestResult]:
    try:
        with zipfile.open(f"{dir}_rels/{file}.rels", "r") as f:
            relationships = ElementTree.parse(f).getroot()
    except KeyError:
        # file does not exist
        return None
    expected_tag = f"{NS_RELATIONSHIPS}Relationships"
    if relationships.tag != expected_tag:
        return AasTestResult(f'Invalid root tag {relationships.tag}, expected {expected_tag}', Level.ERROR)

    if dir:
        result = AasTestResult(f"Checking relationships of {dir}{file}")
    else:
        result = AasTestResult(f"Checking root relationship")
    for idx, rel in enumerate(relationships):
        if rel.tag != f"{NS_RELATIONSHIPS}Relationship":
            result.append(AasTestResult(
                f'Invalid tag {rel.tag}', Level.ERROR))
            continue
        try:
            type = rel.attrib['Type']
            target = rel.attrib['Target']
        except KeyError as e:
            result.append(AasTestResult(f'Attribute {e} is missing', Level.ERROR))
            continue

        if type in DEPRECATED_TYPES:
            new_type = DEPRECATED_TYPES[type]
            result.append(AasTestResult(f"Deprecated type {type}, considering as {new_type}", level=Level.WARNING))
            type = new_type

        if target.startswith('/'):
            target = target[1:]
        else:
            target = dir + target
        target = normpath(target)

        sub_dir, file = splitpath(target)
        sub_rel = Relationship(type, target)
        result.append(AasTestResult(f'Relationship {sub_rel.target} is of type {sub_rel.type}', Level.INFO))
        parent_rel.sub_rels.append(sub_rel)
        if target in visited_targets:
            result.append(AasTestResult(f'Already checked {target}', Level.INFO))
            continue
        visited_targets.add(target)
        if target not in zipfile.namelist():
            result.append(AasTestResult(f'Relationship has non-existing target {target}', Level.ERROR))
            continue
        r = _scan_relationships(zipfile, sub_rel, sub_dir + '/', file, visited_targets)
        if r:
            result.append(r)

    return result


def _check_relationships(zipfile: zipfile.ZipFile, root_rel: Relationship) -> AasTestResult:
    result = AasTestResult('Checking relationships')
    visited_targets = set()
    r = _scan_relationships(zipfile, root_rel, '', '', visited_targets)
    if r:
        result.append(r)
    else:
        result.append(AasTestResult(f"Root relationship does not exist", Level.ERROR))
    return result


def _check_files(zipfile: zipfile.ZipFile, root_rel: Relationship, version: str) -> AasTestResult:
    result = AasTestResult('Checking files')
    origin_rels = root_rel.sub_rels_by_type(TYPE_AASX_ORIGIN)
    if len(origin_rels) != 1:
        result.append(AasTestResult(f"Expected exactly one aas origin, but found {len(origin_rels)}", level=Level.WARNING))
    for aasx_origin in origin_rels:
        spec_rels = aasx_origin.sub_rels_by_type(TYPE_AASX_SPEC)
        if not spec_rels:
            result.append(AasTestResult("No aas spec found", level=Level.WARNING))
        for aasx_spec in spec_rels:
            sub_result = AasTestResult(f'Checking {aasx_spec.target}')
            try:
                with zipfile.open(aasx_spec.target) as f:
                    if aasx_spec.target.endswith('.xml'):
                        r = check_xml_file(f, version)
                    elif aasx_spec.target.endswith('.json'):
                        r = check_json_file(f, version)
                    else:
                        r = AasTestResult('Unknown filetype', Level.WARNING)
                    sub_result.append(r)
            except KeyError:
                return AasTestResult("File does not exist")
            result.append(sub_result)
    return result


def check_aasx_data(zipfile: zipfile.ZipFile, version: str = _DEFAULT_VERSION) -> AasTestResult:

    result = AasTestResult('Checking AASX package')

    result.append(_check_content_type(zipfile))
    if not result.ok():
        return result

    root_rel = Relationship('ROOT', '/')
    result.append(_check_relationships(zipfile, root_rel))
    if not result.ok():
        return result

    result.append(_check_files(zipfile, root_rel, version))
    if not result.ok():
        return result

    return result


def check_aasx_file(file: TextIO, version: str = _DEFAULT_VERSION) -> AasTestResult:
    try:
        zip = zipfile.ZipFile(file)
    except zipfile.BadZipFile as e:
        return AasTestResult(f"Cannot read: {e}", level=Level.ERROR)

    return check_aasx_data(zip, version)


def generate(version: str = _DEFAULT_VERSION) -> Generator[str, None, None]:
    aas = _get_schema(version)
    graph = generate_graph(aas.schema)
    for i in graph.generate_paths():
        sample = graph.execute(i.path)
        if i.is_valid:
            valid = True
        else:
            valid = check_json_data(sample, version).ok()
        yield valid, sample



from typing import Dict, List, Tuple
import base64


def _group(elements: list) -> dict:
    assert isinstance(elements, list)
    grouped: Dict[str, List[any]] = {}
    for item in elements:
        assert isinstance(item, dict)
        key = item.get('idShort')
        assert isinstance(key, str)
        try:
            grouped[key].append(item)
        except KeyError:
            grouped[key] = [item]
    return grouped


def group(data: any) -> any:
    if isinstance(data, list):
        data = [group(i) for i in data]
    elif isinstance(data, dict):
        data: dict = {key: group(value) for key, value in data.items()}
        if data.get('modelType') == 'SubmodelElementCollection':
            elements = data.get('value')
            data['value'] = _group(elements)
        elif data.get('modelType') == 'Submodel':
            elements = data.get('submodelElements')
            data['submodelElements'] = _group(elements)
    return data


def _un_group(elements: dict) -> list:
    if not isinstance(elements, dict):
        return elements
    un_grouped: List[str, dict] = []
    for item in elements.values():
        if isinstance(item, list):
            un_grouped.extend(item)
        else:
            un_grouped.append(item)
    return un_grouped


def un_group(data: any) -> any:
    if isinstance(data, list):
        data = [un_group(i) for i in data]
        for idx, value in enumerate(data):
            data[idx] = un_group(value)
    elif isinstance(data, dict):
        data = {key: un_group(value) for key, value in data.items()}
        if data.get('modelType') == 'SubmodelElementCollection':
            elements = data.get('value')
            data['value'] = _un_group(elements)
        elif data.get('modelType') == 'Submodel':
            elements = data.get('submodelElements')
            data['submodelElements'] = _un_group(elements)
    return data


def normpath(path: str) -> str:
    """
    Normalizes a given path.
    E.g. normpath('///a/../b/)') == '/b'
    This implementation is platform independent and behaves like os.normpath on a unix system.
    See https://docs.python.org/3/library/os.path.html#os.path.normpath for more details.
    """
    path = path.strip()
    if len(path) == 0:
        return ''
    result = []
    for token in path.split("/"):
        if token.strip() == '' or token == '.':
            continue
        if token == '..':
            if result:
                result.pop()
        else:
            result.append(token)
    if path.startswith('/'):
        return "/" + "/".join(result)
    else:
        return "/".join(result)


def splitpath(path: str) -> Tuple[str, str]:
    """
    Splits a path into a pair (head, tail)
    This implementation is platform independent and behaves like os.path.split on a unix system.
    See https://docs.python.org/3/library/os.path.html#os.path.split for more details.
    """
    prefix, _, suffix = path.rpartition('/')
    return prefix, suffix
